//--------------------------------------------------------------------------------------------------
//
// full_duplex_serial_pt.c
//
// A full duplex serial driver using proto threads.
// Runs in a Propeller COG as native code.
//
// Copyright (c) 2011 Michael Rychlik.
// MIT Licensed (see at end of file for exact terms)
//
// History:
//
// 2011-09-26    v0.1  Initial version.
// 2011-10-01    v1.0  Adapted for use as COG driver loadable from LMM code.
//
//--------------------------------------------------------------------------------------------------

#include "full_duplex_serial_pt.h"

// For protothreads.
#include "pt.h"

// For Cog regs (_OUTA, _INA, _DIRA..)
#include "cog.h"

// Pointer to mail box in HUB
volatile struct fds_mailbox* pm;

// Serial port config
static _COGMEM int txPin;
static _COGMEM int rxPin;
static _COGMEM int bitcycles;
static _COGMEM int halfBitcycles;

// tx_Thread variables
static _COGMEM int txWaitCycles;
static _COGMEM int txShiftReg;
static _COGMEM int txBitCnt;

// The UART transmitter protothread function.
// A protothread function must always
// return an integer, but must never explicitly return - returning is
// performed inside the protothread statements.
//
// The protothread function is driven by the main loop further down in
// the code.
_NATIVE static int tx_thread(struct pt *pt)
{
    // A protothread function must begin with PT_BEGIN() which takes a pointer to a struct pt.
    PT_BEGIN(pt);

    // We loop forever here.
    while(1)
    {
        // Wait while the TX FIFO is empty
        PT_WAIT_WHILE(pt, pm->txHead == pm->txTail);

        // Output the start bit
        _OUTA = 0;

        // Load the tx shift register from tx FIFO and add stop bits.
        txShiftReg = pm->txBuffer[(pm->txTail)++ & BUFFER_MASK]; 

        // Get the next bit time
        txWaitCycles = (int)_CNT + bitcycles;

        for (txBitCnt = 0; txBitCnt < 8; txBitCnt++)
        {
            PT_WAIT_UNTIL(pt, (int)_CNT - txWaitCycles >= 0); // Wait for the next bit time
            _OUTA = (txShiftReg & 1) << txPin;                // Output a bit 
            txWaitCycles += bitcycles;                        // Get next bit time
            txShiftReg >>= 1;                                 // Get next bit into position
        }
        // Output the stop bit 
        PT_WAIT_UNTIL(pt, (int)_CNT - txWaitCycles >= 0);     // Wait for the next bit time
        _OUTA = 1 << txPin; 

        // Wait for end  of stop  bit time 
        txWaitCycles += bitcycles;                            // Get next bit time
        PT_WAIT_UNTIL(pt, (int)_CNT - txWaitCycles >= 0);
    }
    // All protothread functions must end with PT_END() which takes a pointer to a struct pt.
    PT_END(pt);
}

// rx_thread variables
static _COGMEM int rxWaitCycles;
static _COGMEM int rxShiftReg;
static _COGMEM int rxBitCnt;
static _COGMEM int bit;

// The UART reciever protothread function.
// A protothread function must always
// return an integer, but must never explicitly return - returning is
// performed inside the protothread statements.
//
// The protothread function is driven by the main loop further down in
// the code.
_NATIVE static int rx_thread(struct pt *pt)
{
    // A protothread function must begin with PT_BEGIN() which takes a pointer to a struct pt.
    PT_BEGIN(pt);

    // We loop forever here.
    while(1)
    {
        rxShiftReg = 0;
        // Wait until a start bit is detected
        PT_WAIT_UNTIL(pt, ((_INA >> rxPin) & 1) == 0);

        // Wait until halfway through start bit
        rxWaitCycles = (int)_CNT + halfBitcycles;

        PT_WAIT_UNTIL(pt, (int)_CNT - rxWaitCycles >= 0);
        rxWaitCycles = (int)_CNT + bitcycles;                   // Get next bit time
        
        for (rxBitCnt = 0; rxBitCnt <= 7; rxBitCnt++)
        { 
            PT_WAIT_UNTIL(pt, (int)_CNT - rxWaitCycles >= 0);   // Wait for middle of bit time 
            bit = (_INA >> rxPin) & 1;                          // Read the bit 
            rxWaitCycles = (int)_CNT + bitcycles;               // Get next bit time               
            rxShiftReg |= bit << rxBitCnt;                      // and place in shift register
        } 

	// Write the recieved char to the RX FIFO
	if ((pm->rxHead - pm->rxTail) < BUFFER_SIZE - 1)
        {
                // Not full: Put char in buffer.
                pm->rxBuffer[pm->rxHead++ & BUFFER_MASK] = (char)rxShiftReg;
	}

        // Wait into center of Stop bit
        PT_WAIT_UNTIL(pt, (int)_CNT - rxWaitCycles >= 0); 
    }

    // All protothread functions must end with PT_END() which takes a pointer to a struct pt.
    PT_END(pt);
}

static _COGMEM struct pt* pPtTx;
static _COGMEM struct pt* pPtRx;

// Finally, we have the main loop.
// Here is where the protothreads are initialized and scheduled.
_NATIVE void main (volatile struct fds_mailbox *m)
{
    // Copy mail box pointer for other functions use.
    pm = m;
    pPtTx = &pm->pt_tx;
    pPtRx = &pm->pt_rx;

    // Set up serial port
    rxPin = pm->rxPin;
    txPin = pm->txPin;
    bitcycles = 80000000 / pm->baudRate;  //<HEATER>Shouldn't CLKFREQ be defined somewhere visible here?</HEATER>
    halfBitcycles = bitcycles / 2;
    _OUTA = (unsigned)1 << txPin;
    _DIRA = (unsigned)1 << txPin;

    // Initialize the protothread state variables with PT_INIT().
    PT_INIT(pPtTx);
    PT_INIT(pPtRx);

    // Schedule the two protothreads by repeatedly calling their
    // protothread functions and passing a pointer to the protothread
    // state variables as arguments.
    while(1)
    {
        // Run the transmiter thread a while
        tx_thread (pPtTx);

        // Run the reciever thread a while
        rx_thread (pPtRx);
    }
}

//--------------------------------------------------------------------------------------------------
// TERMS OF USE: MIT License
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software,
// and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//--------------------------------------------------------------------------------------------------
